<?php

/**
 * @file
 * FFmpeg wrapper class. Provides the core functionality for handling video
 * files.
 * @todo Implement file naming on random basis to make file names unique and
 *   avoid possible naming clashes.
 */
class video_handler_ffmpeg {
  public $duration;
  private $filename;
  private $filext;
  private $filedir;
  private $file_info;
  private $duration_in_seconds;
  private $resolution;
  private $width;
  private $height;
  private $aspect_ratio;
  private $video_codec;
  private $audio_codec;
  private $sampling_rate;
  private $audio_mode;
  private $audio_bitrate;
  private $target_dir;
  private $node;
  private $filename_full;
  private $file;
  private $ffmpeg_dir;
  private $private_files;
  private $transcode_options;
  private $scale_width;
  private $thumb_width;
  private $thumb_height;
  private $preview_options;
  private $video_field;
  private $thumb_field;
  private $transcoded_video_field;
  private $job_status;

  /**
   * @param int $nid
   *   Node object, which is being processed.
   * @param string $filename
   *   Name of the file to be processed including its extension, e.g. file.ext.
   * @param string $filepath
   *   Drupal relative filepath to the file.
   * @global object $user
   *   Drupal's $user object.
   */
  public function __construct($node, $filename, $filepath) {
    global $user;
    $this->node = $node;
    $this->filename_full = $filename;
    $this->filename = substr($this->filename_full, 0, strpos($this->filename_full, '.'));
    $this->filext = substr($this->filename_full, strrpos($this->filename_full, '.') + 1);
    $this->file = $_SERVER['DOCUMENT_ROOT'] . '/' . $filepath;
    $this->filedir = substr($this->file, 0, strpos($this->file, $this->filename_full) - 1);
    $this->private_files = variable_get('vh_private_files', FALSE);
    // If private files directory is enabled, we need to create target directory
    // for transcoded file(s).
    if ($this->private_files) {
      // Path to directory where transcoded file(s) will be placed to.
      $this->target_dir = $_SERVER['DOCUMENT_ROOT'] . '/' . file_directory_path() .
                          '/private/' . $user->uid . '/' . $this->node->nid;
      // Create this directory.
      mkdir($this->target_dir, 0744, TRUE);
    }
    else {
      $this->target_dir = $this->filedir;
    }
    $this->ffmpeg_dir = variable_get('vh_ffmpeg_dir', '');
    $this->transcode_options = variable_get('vh_transcode_options', '');
    $this->file_info = $this->get_file_info();
    $this->duration = $this->get_file_info_detail($this->file_info, 'duration');
    $this->duration_in_seconds = $this->get_duration_in_seconds($this->duration);
    $this->resolution = $this->get_file_info_detail($this->file_info, 'resolution');
    $dimensions = explode('x', $this->resolution);
    $this->width = $dimensions[0];
    $this->height = $dimensions[1];
    $this->aspect_ratio = $this->width / $this->height;
    $this->video_codec = $this->get_file_info_detail($this->file_info, 'video_codec');
    $this->audio_codec = $this->get_file_info_detail($this->file_info, 'audio_codec');
    $this->sampling_rate = $this->get_file_info_detail($this->file_info, 'sampling_rate');
    $this->audio_mode = $this->get_file_info_detail($this->file_info, 'audio_mode');
    $this->audio_bitrate = $this->get_file_info_detail($this->file_info, 'audio_bitrate');
    $this->scale_width = $this->transcode_options['scale_width'];
    $this->thumb_width = $this->transcode_options['thumb_width'];
    $this->thumb_height = round($this->thumb_width / $this->aspect_ratio);
    $this->preview_options = variable_get('vh_preview_options', '');
    $fields_to_handle = variable_get('vh_fields_to_handle', '');
    if ($fields_to_handle) {
      $this->video_field = $fields_to_handle['video_field'];
      $this->thumb_field = $fields_to_handle['thumb_field'];
      $this->transcoded_video_field = $fields_to_handle['transcoded_video_field'];
    }
  }
  
  /**
   * Runs FFmpeg with required parameters.
   * 
   * @param string $params
   *   Parameters to run FFmpeg with.
   * @param bool $output
   *   Indicates whether or not turn on buffer output.
   * @param string $preinput
   *   Parameter(s) to pass before -i parameter.
   * @return
   *   The contents of the output buffer is returned, if $output parameter
   *   evaluates to TRUE, otherwise nothing is returned.
   */
  private function run_command($params = '', $output = FALSE, $preinput = '') {
    $redirect_output = ($output) ? ' 2>&1' : '';
    $fileinput = escapeshellarg($this->file);
    $command = $this->ffmpeg_dir .  '/ffmpeg' . $preinput . ' -i ' . $fileinput .
               $params . $redirect_output;
    ob_start();
    passthru($command);
    if ($output) {
      $out = ob_get_contents();
    }
    ob_end_clean();
    if ($output) {
      return $out;
    }
  }
  
  /**
   * Transcodes videos. Main workhorse method.
   * 
   * @global object $user
   *   Drupal's $user object.
   */
  public function transcode() {
    global $user;
    $audio_params = $this->audio_params();
    $i = 0;
    foreach ($this->scale_width as $scale_width) {
      $frame_size = $this->frame_size($scale_width);
      // If uploaded file meets all the requirements it doesn't need to be
      // transcoded. Otherwise, build parameters string accordigly to pass to
      // transcoder.
      if ($this->video_codec == 'h264' && $frame_size == $this->resolution &&
          $this->filext == 'mp4' && $audio_params == ' -acodec copy') {
        $transcode_params = FALSE;
      }
      else {
        $video_params = $this->video_params($frame_size);
        $transcoded_filename = $this->filename . '_' . $frame_size . '.mp4';
        // The path for output file.
        $fileoutput = $this->target_dir . '/' . $transcoded_filename;
        // Make it safe for passing to shell command as it contains the user
        // input.
        $fileoutput_safe = escapeshellarg($fileoutput);
        $transcoded_filepath = substr($fileoutput, strpos($fileoutput, file_directory_path()));
        $transcode_params = $video_params . $audio_params . ' ' . $fileoutput_safe;
      }
      
      // Transcode the video if needed, and if not - just append frame size to
      // the name of the uploaded file and copy it to the private files
      // directory if it's specified, or leave it in the same directory
      // otherwise.
      if ($transcode_params) {
        // Perform transcoding.
        $this->run_command($transcode_params);
        // Build the file object and perform required operations on it and
        // regarding the transcoded file.
        if (file_exists($fileoutput)) {
          $node = $this->node;
          $file = new stdClass();
          $file->filename = $transcoded_filename;
          $file->filepath = $transcoded_filepath;
          $file->filemime = file_get_mimetype($transcoded_filename);
          $file->filesize = filesize($fileoutput);
          $file->timestamp = time();
          // If private file system is disabled we need to place the record
          // about the first transcoded file instead of original one in the
          // database, because original file will be deleted at the end of all
          // the process.
          // Otherwise, make record regarding created file to the database and
          // add that file to the node in question.
          if (!$this->private_files && $i == 0) {
            $file->fid = $node->{$this->video_field}[0]['fid'];
            drupal_write_record('files', &$file, array('fid'));
          }
          else {
            $file->uid = $user->uid;
            $file->status = FILE_STATUS_PERMANENT;
            drupal_write_record('files', &$file);
            $node->{$this->transcoded_video_field}[] = field_file_load($transcoded_filepath);
            node_save($node);
          }
        }
      }
      else {
        // Build the path for destination file.
        $target_filename = $this->filename . '_' . $this->resolution . '.mp4';
        $dest = $this->target_dir . '/' . $target_filename;
        $target_filepath = substr($dest, strpos($dest, file_directory_path()));
        copy($this->file, $dest);
        if ($i == 0) {
          // Place the record about the copied file instead of original one in
          // the database, because original file will be deleted at the end of
          // all the process.
          if (file_exists($dest)) {
            $node = $this->node;
            $file = new stdClass();
            $file->fid = $node->{$this->video_field}[0]['fid'];
            $file->filename = $target_filename;
            $file->filepath = $target_filepath;
            $file->filemime = file_get_mimetype($target_filename);
            drupal_write_record('files', &$file, array('fid'));
          }
        }
      }
      $i++;
    }
    // Update job status indicating the stage of the process.
    $this->job_status = 1;
    // Generate thumbnail preview image.
    $this->make_thumb();
    // If private files directory is used, we need to create a preview. Then
    // check for job status and delete original uploaded file.
    if ($this->private_files) {
      $this->make_preview();
      if ($this->job_status == 3) {
        unlink($this->file);
      }
    }
    else {
      if ($this->job_status == 2) {
        unlink($this->file);
      }
    }
  }
  
  /**
   * Determines whether or not audio track is required to be transcoded, and if
   * so, builds parameters string accordingly.
   * 
   * @return string
   *   String representing audio encoding parameters to pass to transcoder.
   */
  private function audio_params() {
    $options = $this->transcode_options;
    $abitrate = ($this->audio_bitrate > $options['abitrate']) ?
                ' -ab ' . $options['abitrate'] . 'k' : '';
    $samplingrate = ($this->sampling_rate > $options['samplingrate']) ?
                    ' -ar ' . $options['samplingrate'] : '';
    $aac_allowed_modes = array('mono', 'stereo');
    $achannels = ($this->audio_mode == '1 channels' || $this->audio_mode == 'mono') ? ' -ac 1' : ' -ac 2';
    if ($this->audio_codec == 'aac' && !$abitrate && !$samplingrate &&
        in_array($this->audio_mode, $aac_allowed_modes)) {
      $audio_params = ' -acodec copy';
    }
    else {
      $acodec = ' -acodec libfaac' . $achannels;
      $audio_params = $acodec . $abitrate . $samplingrate;
    }
    return $audio_params;
  }
  
  /**
   * Determines whether or not video is required to be transcoded, and if so,
   * builds parameters string accordingly.
   * 
   * @param string $frame_size
   *   Frame size to use.
   * @return string
   *   String representing video encoding parameters to pass to transcoder.
   */
  private function video_params($frame_size) {
    $options = $this->transcode_options;
    if ($this->video_codec == 'h264' && $frame_size == $this->resolution
        && $this->filext == 'mp4') {
      $video_params = ' -vcodec copy';
    }
    else {
      $vcodec = ' -vcodec libx264';
      $preset = ' -preset ' . $options['preset'];
      $crf = ' -crf ' . $options['crf'];
      $s = ' -s ' . $frame_size;
      $threads = ' -threads ' . $options['threads'];
      $video_params = $vcodec . $preset . $crf . $s . $threads;
    }
    return $video_params;
  }
  
  /**
   * Calculates frame size for video transcoding.
   * 
   * @param string $scale_width
   *   Width to use for calculation.
   * @return string
   *   Frame size in 640x480 like format.
   */
  private function frame_size($scale_width) {
    if ($this->width > $scale_width) {
      $width = $scale_width;
      $height = ceil($width / $this->aspect_ratio);
      // Height must be even, otherwise it'll cause an error and file won't be
      // transcoded.
      $height = ($height % 2 == 0) ? $height : $height - 1;
      $frame_size = $width . 'x' . $height;
    }
    else {
      $frame_size = $this->resolution;
    }
    return $frame_size;
  }
  
  /**
   * Builds thumbnail generation command parameters and runs the command with
   * these parameters.
   * 
   * @global object $user
   *   Drupal's $user object.
   */
  private function make_thumb() {
    global $user;
    // Frame size parameter.
    $s = ' -s ' . $this->thumb_width . 'x' . $this->thumb_height;
    // Parameter of time position in the video at which thumb image will be
    // taken (it will be put in front of the input in the ffmpeg command).
    $ss = ' -ss ' . round($this->duration_in_seconds / 2);
    // Constructing thumb image full filepath.
    $thumb_filename = $this->filename . '.jpg';
    $thumboutput = $this->filedir . '/' . $thumb_filename;
    $thumboutput_safe = escapeshellarg($thumboutput);
    $thumb_filepath = substr($thumboutput, strpos($thumboutput, file_directory_path()));
    $thumb_params = $ss . ' -vframes 1 -f image2' . $s . ' ' . $thumboutput_safe;
    // Generating thumbnail image.
    $this->run_command($thumb_params);
    if (file_exists($thumboutput)) {
      $file = new stdClass();
      $file->uid = $user->uid;
      $file->status = FILE_STATUS_PERMANENT;
      $file->filename = $thumb_filename;
      $file->filepath = $thumb_filepath;
      $file->filemime = file_get_mimetype($thumb_filename);
      $file->filesize = filesize($thumboutput);
      $file->timestamp = time();
      drupal_write_record('files', &$file);
      $node = $this->node;
      $node->{$this->thumb_field}[0] = field_file_load($thumb_filepath);
      node_save($node);
      $this->job_status++;
    }
  }
  
  /**
   * Makes short preview from uploaded video file.
   */
  private function make_preview() {
    $frame_size = $this->frame_size($this->preview_options['width']);
    $video_params = $this->video_params($frame_size);
    $audio_params = $this->audio_params();
    if ($this->preview_options['offset'] == 0) {
      $ss = 0;
    }
    elseif ($this->preview_options['offset'] == 1) {
      $ss = round(($this->duration_in_seconds - $this->preview_options['length']) / 2);
    }
    $preinput = ' -ss ' . $ss . ' -t ' . $this->preview_options['length'];
    $preview_filename = $this->filename . '_preview.mp4';
    $previewoutput = $this->filedir . '/' . $preview_filename;
    $previewoutput_safe = escapeshellarg($previewoutput);
    $preview_filepath = substr($previewoutput, strpos($previewoutput, file_directory_path()));
    $preview_params = $video_params . $audio_params. ' ' . $previewoutput_safe;
    $this->run_command($preview_params, FALSE, $preinput);
    if (file_exists($previewoutput)) {
      $node = $this->node;
      $file = new stdClass();
      $file->fid = $node->{$this->video_field}[0]['fid'];
      $file->filename = $preview_filename;
      $file->filepath = $preview_filepath;
      $file->filemime = file_get_mimetype($preview_filename);
      $file->filesize = filesize($previewoutput);
      $file->timestamp = time();
      drupal_write_record('files', &$file, array('fid'));
      $this->job_status++;
    }
  }
  
  /**
   * Get detailed info about uploaded video file.
   */
  private function get_file_info() {
    return $this->run_command('', TRUE);
  }
  
  /**
   * Extracts specific piece of info about the video file.
   * 
   * @param string $file_info
   *   Full info about the file.
   * @param string $detail
   *   Type of info to extract.
   */
  private function get_file_info_detail($file_info, $detail = '') {
    switch ($detail) {
      case 'resolution':
        $search = '/Video: (.*?), (.*?), (.*?)( .*?)*,/';
        $i = 3;
        break;
      case 'duration':
        $search = '/Duration: (.*?),/';
        $i = 1;
        break;
      case 'video_codec':
        $search = '/Video: (.*?)( .*?)*,/';
        $i = 1;
        break;
      case 'audio_codec':
        $search = '/Audio: (.*?)( .*?)*,/';
        $i = 1;
        break;
      case 'sampling_rate':
        $search = '/Audio: (.*?), (.*?) Hz,/';
        $i = 2;
        break;
      case 'audio_mode':
        $search = '/Audio: (.*?), (.*?), (.*?),/';
        $i = 3;
        break;
      case 'audio_bitrate':
        $search = '/Audio: (.*?), (.*?), (.*?), (.*?), (.*?) kb\/s/';
        $i = 5;
        break;
    }
    $matches = array();
    preg_match($search, $file_info, &$matches, PREG_OFFSET_CAPTURE);
    return $matches[$i][0];
  }
  
  /**
   * Calculates duration of video in seconds only.
   * 
   * @param string $hms
   *   Duration in hh:mm:ss.ms format.
   * @return string
   *   Duration in seconds.
   */
  private function get_duration_in_seconds($hms) {
    // Cutting off milliseconds.
    $hms = substr($hms, 0, strpos($hms, '.'));
    list($h, $m, $s) = explode(':', $hms);
    $seconds = 0;
    $seconds += intval($h) * 3600;
    $seconds += intval($m) * 60;
    $seconds += intval($s);
    return $seconds;
  }
}